"""Functions to read configuration.
"""
import configparser
import logging
import os
import sys
from glob import glob
from typing import List, Optional

_logger = logging.getLogger(__name__)


class Config(object):
    """Singleton for the configuration.
    """

    __instance = None

    def __new__(cls):
        if Config.__instance is None:
            Config.__instance = Configuration()
        return Config.__instance


class Configuration(object):
    """
    At the moment, only read from `setup.cfg`.
    """

    def __init__(self, base_path: str = None):
        self._expanded_configuration = dict()
        setup_path = "setup.cfg"
        if base_path:
            setup_path = os.path.join(base_path, setup_path)
        section = "odoo_scripts"
        config_parser = configparser.ConfigParser()
        if not os.path.exists(setup_path):
            _logger.warning("Missing %s", setup_path)
        else:
            config_parser.read(setup_path)

        def set_from_glob_values(
            base: Optional[str], glob_values: List[str]
        ) -> set:
            result_set = set()
            for glob_value in glob_values:
                if base:
                    glob_value = os.path.join(base, glob_value)
                for file in glob(glob_value):
                    if os.path.isdir(file):
                        result_set.add(file)
            return result_set

        def get_expanded_configuration(path_name: str):
            if path_name not in self._expanded_configuration:
                self._expanded_configuration[path_name] = Configuration(
                    path_name
                )
            return self._expanded_configuration[path_name]

        for key in ("modules", "dependencies", "other_sources"):
            # get the values from this configuration
            values = config_parser.get(section, key, fallback="").split()
            set_values = set_from_glob_values(base_path, values)
            # handle expand, getting value from another configuration
            expand_value = config_parser.get(
                section, key + ".expand", fallback=None
            )
            if expand_value:
                for element in expand_value.split():
                    set_values.update(
                        getattr(get_expanded_configuration(element), key)
                    )
            setattr(self, key, list(set_values))

        def key_format(a_key: str) -> str:
            """Replace invalid characters in an attribute"""
            return a_key.replace("-", "_").replace(".", "_")

        for key in ("module_list", "module_list_tests", "pg.extensions"):
            value = list()
            formatted_key = key_format(key)
            # add any expanded value
            expand_value = config_parser.get(
                section, key + ".expand", fallback=None
            )
            if expand_value:
                for element in expand_value.split():
                    value.extend(
                        getattr(
                            get_expanded_configuration(element), formatted_key
                        )
                    )
            # add local values
            value += config_parser.get(section, key, fallback="").split()
            # then set it
            setattr(self, formatted_key, value)
        for key in ("db_user", "db_password", "load-language"):
            setattr(
                self,
                key_format(key),
                config_parser.get(section, key, fallback=None),
            )

        self.registry = config_parser.get(
            section, "registry", fallback="registry.xcg.io"
        )
        project_path = os.path.realpath(".")
        self.image = config_parser.get(
            section, "image", fallback=os.path.basename(project_path)
        )
        self.odoo_type = config_parser.get(
            section, "odoo_type", fallback="odoo7"
        )
        if self.odoo_type not in (
            "odoo7",
            "odoo8",
            "odoo10",
            "odoo11",
            "odoo13",
        ):
            _logger.warning("Unexpected odoo_type: %s", self.odoo_type)
        self.postgresql_version = config_parser.get(
            section, "postgresql_version", fallback="9.6"
        )
        self.start_py3o = config_parser.get(
            section, "start_py3o", fallback="no"
        ) in ("yes", "true")


def main(argv=None):
    """Display configuration. Useful for development of odoo_scripts.
    """
    c = Config()
    for key, value in c.__dict__.items():
        print(key, "=", value)
    return 0


if __name__ == "__main__":
    return_code = main(sys.argv[1:])
    if return_code:
        exit(return_code)
