#!/usr/bin/env python3
# vim: set shiftwidth=4 softtabstop=4:
"""Script to locally build a docker image
"""
import datetime
import json
import logging
import os
import signal
import sys
from subprocess import call, check_output

from .config import Config
from .docker_build_clean import clean
from .docker_build_copy import copy
from .docker_client import DockerClient
from .parsing import apply, basic_parser

_logger = logging.getLogger(__name__)

__version__ = "1.0.0"
__date__ = "2018-04-04"
__updated__ = "2020-09-11"


def __parser():
    program_version = __version__
    program_build_date = str(__updated__)
    program_version_message = "%%(prog)s %s (%s)" % (
        program_version,
        program_build_date,
    )
    program_shortdesc = __doc__.split(".")[0]
    program_license = """%s

      Created by Vincent Hatakeyama on %s.
      Copyright 2018, 2019, 2020 XCG Consulting. All rights reserved.

      Licensed under the MIT License

      Distributed on an "AS IS" basis without warranties
      or conditions of any kind, either express or implied.

    USAGE
    """ % (
        program_shortdesc,
        str(__date__),
    )
    parser = basic_parser(program_license, program_version_message)
    return parser


def main(argv=None):  # IGNORE:C0111
    """Parse arguments and docker build
    """
    # TODO the script assume it is launched from the parent super project
    project_path = os.path.realpath(".")
    project_name = os.path.basename(project_path)
    if project_name == "odoo_scripts":
        _logger.fatal("You must run this script from the super project")
        return 1

    # TODO add a way to store configuration options in a project file
    # Argument parsing
    parser = __parser()
    # TODO add tag option, and maybe force tag option
    parser.add_argument(
        "--ensureconf",
        help="ensureconf [default: %(default)s]",
        action="store_true",
    )
    parser.add_argument(
        "--push", help="Push image [default: %(default)s]", action="store_true"
    )
    parser.add_argument(
        "--dev",
        help="add dev feature to generated image [default: %(default)s]",
        action="store_true",
    )
    parser.add_argument(
        "--build-arg",
        help="build arg for the image, formated like FOO=BAR "
        "[default: %(default)s]",
        default=None,
        nargs="*",
    )
    parser.add_argument(
        "--no-pull",
        help="indicate to docker to not pull the base image "
        "[default: %(default)s]",
        action="store_true",
    )
    # TODO (maybe) add argument for other build arg

    nmspc = parser.parse_args(argv)
    apply(nmspc)
    ensureconf = nmspc.ensureconf
    dev = nmspc.dev
    push = nmspc.push

    c = Config()
    registry = c.registry
    project = c.image
    odoo_type = c.odoo_type
    image = "%s/%s:latest" % (registry, project)
    _logger.debug("Docker image: %s", image)
    # TODO ensureconf
    if ensureconf:
        raise NotImplementedError
    # call build copy
    copy()

    # clean on exit
    def signal_handler(code, frame):
        # TODO handle the case when odoo_scripts is not installed
        clean()
        # XXX needed?

    signal.signal(signal.SIGINT, signal_handler)
    signal.signal(signal.SIGTERM, signal_handler)

    buildargs = dict()
    if os.path.exists(".hg"):
        tags = check_output(["hg", "identify", "--tags"]).decode()
        if tags:
            buildargs["VERSION"] = tags
        buildargs["VCS_URL"] = check_output(
            ["hg", "paths", "default"]
        ).decode()
        buildargs["VCS_REF"] = check_output(
            ["hg", "identify", "--id"]
        ).decode()
    buildargs["BUILD_DATE"] = datetime.datetime.now().isoformat()
    if nmspc.build_arg:
        for arg in nmspc.build_arg:
            a = arg.split("=")
            buildargs[a[0]] = a[1]
    _logger.debug("Build args: %s", buildargs)
    dockerfile = "Dockerfile"
    if dev:
        debug_dockerfile = "Dockerfile.debug"
        call(["cp", dockerfile, debug_dockerfile])

        with open(debug_dockerfile, "a") as myfile:
            myfile.write("\n# Developer helpers\n" "RUN apt-get update -qq\n")
            myfile.write(
                "RUN DEBIAN_FRONTEND=noninteractive apt-get install -y -qq "
            )
            if odoo_type in ("odoo11", "odoo13"):
                myfile.write(
                    "python3-watchdog python3-ipdb python3-pyinotify\n"
                )
                # Install pdb++
                myfile.write("RUN pip3 install pdbpp\n")
            elif odoo_type in ("odoo10",):
                myfile.write("python-ipdb\n")
                myfile.write(
                    "RUN pip install watchdog --disable-pip-version-check "
                    "--system --no-cache-dir --only-binary wheel"
                )
            elif odoo_type in ("odoo7", "odoo8"):
                myfile.write("python-ipdb\n")
                myfile.write("RUN pip install pyinotify")
        dockerfile = debug_dockerfile
        # TODO remove temp image

    docker_client = DockerClient.client
    pull = not nmspc.no_pull
    _logger.debug("Docker Pull %s", pull)
    builder = docker_client.api.build(
        path=".",
        rm=True,
        pull=pull,
        buildargs=buildargs,
        tag=image,
        dockerfile=dockerfile,
    )
    for line in builder:
        d = json.loads(line.decode("utf-8"))
        if "stream" in d:
            for line in d["stream"].rstrip().split("\n"):
                _logger.info(line)
        if "errorDetail" in d:
            _logger.fatal(d["errorDetail"])
            return 1
    if dev:
        call(["rm", dockerfile])

    # TODO docker tag with tags/bookmarks (unused so maybe no need)
    # TODO docker push (only when asked for)
    if push:
        raise NotImplementedError
    # XXX call cleanup more intelligently
    signal_handler(0, None)
    return 0


if __name__ == "__main__":
    return_code = main(sys.argv[1:])
    if return_code:
        exit(return_code)
