import logging
import os
import tarfile
from typing import Dict, List

import docker
import dockerpty

_logger = logging.getLogger(__name__)
if docker.__version__ < "3.4.0":
    _logger.warning(
        "Unexepected python docker version: %s", docker.__version__
    )

# TODO detect that user is member of docker group


class DockerClient:
    client = docker.from_env()

    @classmethod
    def create_volume(
        cls, volume_name: str, extra_labels: Dict[str, str] = None
    ) -> docker.models.volumes.Volume:
        """Return the volume passed in parameter, creating it if it does not
        exists.
        :param volume_name: name of the volume to create
        :param extra_labels: extra labels to put on the volume (only on
        creation)
        """
        volumes = cls.client.volumes.list(filters={"name": volume_name})
        if volumes:
            _logger.debug("Volume %s already exists", volume_name)
            return volumes[0]
        else:
            _logger.debug("Creating volume %s", volume_name)
            labels = dict(odoo_scripts="")
            if extra_labels:
                labels.update(extra_labels)
            return cls.client.volumes.create(name=volume_name, labels=labels)

    @staticmethod
    def put_file(src, container, dst):
        # copied from https://stackoverflow.com/questions/46390309/how-to-copy-a-file-from-host-to-container-using-docker-py-docker-sdk # noqa: E501
        os.chdir(os.path.dirname(src))
        srcname = os.path.basename(src)
        # TODO put that in a temporary directory
        tar = tarfile.open(src + ".tar", mode="w")
        try:
            tar.add(srcname)
        finally:
            tar.close()

        data = open(src + ".tar", "rb").read()
        container.put_archive(os.path.dirname(dst), data)

    @classmethod
    def exec(
        cls,
        container: docker.models.containers.Container,
        command: List[str],
        user: str,
        interactive: bool = False,
    ) -> int:
        """Execute a command in a container

        :param container: Container to exec on
        :param command: Command to be executed
        :param user: User to execute command as.
        :param interactive: True if the container needs to be interactive
        """
        if interactive:
            dockerpty.exec_command(cls.client.api, container.id, command)
            # it would be nice to return the exec command return code, but
            # there is no way to do that apparently
            result = 0
        else:
            run_result = container.exec_run(command, user=user)
            result = run_result.exit_code
            print(run_result.output.decode())
        return result
